"""Utility Functions for Supporting Other Modules"""
import requests
import json
import time
from servicepytan import URL_ROOT, AUTH_ROOT
from servicepytan.auth import get_auth_headers, get_tenant_id

def request_json(url, options={}, payload="", conn=None, request_type="GET", json_payload=""):
  """Makes the request to the API and returns JSON

  Retrieves JSON response from provided URL with a number of parameters to customize the request.

  Args:
      url: A string with the full URL request
      options: A dictionary defining the parameters to add to the url for filtering
      payload: A dictionary defining the data object to create or update
      conn: a dictionary containing the credential config.
      request_type: A string to define the REST endpoint type [GET, POST, PUT, PATCH, DEL].

  Returns:
      JSON Object

  Raises:
      TBD
  """
  headers = get_auth_headers(conn)
  response = requests.request(request_type, url, data=payload, headers=headers, params=options, json=json_payload)
  return json.loads(response.text)

def check_default_options(options):
  """Add sensible defaults to options when not defined"""
  # TODO: Add ability to read from a configuration file
  if "pageSize" not in options:
    options["pageSize"] = 100
  
  return options

def endpoint_url(folder, endpoint, id="", modifier="", conn=None, tenant_id=""):
  """Constructs API request URL based on key parameters

  Retrives JSON response from provided URL with a number of parameters to customize the request.

  Args:
      folder: A string based on the endpoint groupings
      endpoint: A string indicating the endpoint you want to address.
      id: A string for the id of the endpoint object you're addressing.
      modifier: A string to modify the url to address the additional endpoint.
      conn: a dictionary containing the credential config.
      tenant_id: A string to manually adjust the tenant id.

  Returns:
      A URL String

  Raises:
      TBD
  """  
  # Adds ability to manually switch up the Tenant ID for apps that have multiples
  if tenant_id == "":
    tenant_id = get_tenant_id(conn)

  url = f"{URL_ROOT}/{folder}/v2/tenant/{tenant_id}/{endpoint}"
  if id != "": url = f"{url}/{id}"
  if modifier != "": url = f"{url}/{modifier}"
  return url
    
def create_credential_file(name="servicepytan_config.json"):
  """Creates and saves an unfilled configuration file.

  Args:
      name: a dictionary containing the credential config

  Returns:
      A filepath string
  """  
  file = open(name, 'w')
  file.write(
    """{
    "SERVICETITAN_CLIENT_ID": "",
    "SERVICETITAN_CLIENT_SECRET": "",
    "SERVICETITAN_APP_ID": "",
    "SERVICETITAN_APP_KEY": "",
    "SERVICETITAN_TENANT_ID": ""
    }"""
  )
  file.close()
  return name

def get_timezone_by_file(conn=None):
  """Retrieves timezone from the configuration file.

  Args:
      conn: a dictionary containing the credential config

  Returns:
      Timezone string
  """    
  # Read File
  if "SERVICETITAN_TIMEZONE" in conn:
    timezone = config['SERVICETITAN_TIMEZONE']
  else:
    timezone = "UTC"
  return timezone

def sleep_with_countdown(sleep_time):
  """Sleeps for a given amount of time with a countdown"""
  for i in range(sleep_time, 0, -1):
      print("Trying again in {} seconds...       ".format(i),end='\r')
      time.sleep(1)
  print("")
  pass

def request_json_with_retry(url, options={}, payload="", conn=None, request_type="GET", json_payload=""):
  """Makes the request to the API and returns JSON with a retry

  Retrieves JSON response from provided URL with a number of parameters to customize the request.

  Args:
      url: A string with the full URL request
      options: A dictionary defining the parameters to add to the url for filtering
      payload: A dictionary defining the data object to create or update
      conn: a dictionary containing the credential config.
      request_type: A string to define the REST endpoint type [GET, POST, PUT, PATCH, DEL].
      retry_count: An integer for the number of times to retry the request.
      sleep_time: An integer for the number of seconds to sleep between retries.

  Returns:
      JSON Object

  Raises:
      TBD
  """
  response = request_json(url, options=options, payload=payload, conn=conn, request_type=request_type, json_payload=json_payload)
  if "traceId" in response:
    if response['status'] == 429:
        sleep_time = response['title'].split(" ")[-2]
        print("Rate Limit Exceeded. Retrying in {} seconds...".format(sleep_time))
        sleep_with_countdown(int(sleep_time))
        response = request_json_with_retry(url, options=options, payload=payload, conn=conn, request_type=request_type, json_payload=json_payload)
  
  return response


from servicepytan.data import DataService

def get_booked_jobs_by_agent(start_date, end_date, conn=None):
  booked_jobs = DataService(conn=conn).get_jobs_created_between(start_date, end_date)
  employees = DataService(conn=conn).get_employees()
  pass2


from servicepytan.utils import request_json, check_default_options, endpoint_url
class Endpoint:
  """Primary class for interacting with the API by establishing an endpoint object.

  The core way of getting and changing data in ServiceTitan using the library is to create
  an endpoint object. Each method pulls the data in a different way and may or may not apply 
  to a certain endpoint. You will need to consult the developer docs to identify the exact
  parameters needed for a given endpoint and which methods will apply.

  Attributes:
      folder: A string indicating the group of endpoints you want to address.
      endpoint: A string indicating the endpoint you want to address.
      conn: a dictionary containing the credential config.
  """
  def __init__(self, folder, endpoint, conn=None):
    """Inits Endpoint with folder, endpoint and allows for getting necessary credentials from the config file."""
    self.folder = folder
    self.endpoint = endpoint
    self.conn = conn

  # Main Request Types
  def get_one(self, id, modifier=""):
    """Retrieve one record using the record id. Modifier is used for further endpoints."""
    url = endpoint_url(self.folder, self.endpoint, id=id, modifier=modifier, conn=self.conn)
    return request_json(url, options={}, payload="", conn=self.conn, request_type="GET")

  def get_many(self, query={}):
    """Retrieve one page of results with query options to customize."""
    url = endpoint_url(self.folder, self.endpoint, conn=self.conn)
    options = check_default_options(query)
    return request_json(url, options, payload="", conn=self.conn, request_type="GET")
  
  def get_all(self, query={}):
    """Retrive all pages in your query."""
    query["page"] = "1"
    print(query)
    response = self.get_many(query)
    data = response["data"]
    if data == []: return []
    has_more = response["hasMore"]
    while has_more:
      query["page"] = str(int(query["page"]) + 1)
      print(query)
      response = self.get_many(query)
      data.extend(response["data"])
      has_more = response["hasMore"]

    return data

  def create(self, payload):
    """Method that corresponds to a POST request for creating objects"""
    url = endpoint_url(self.folder, self.endpoint, conn=self.conn)
    return request_json(url, options={}, payload=payload, conn=self.conn, request_type="POST")

  def update(self, id, payload, modifier="", request_type="PUT"):
    """Method that corresponds to PUT or POST request for updating objects. Defaults to PUT"""
    url = endpoint_url(self.folder, self.endpoint, id=id, modifier=modifier, conn=self.conn)
    return request_json(url, options={}, payload=payload, conn=self.conn, request_type=request_type)

  def delete(self, id, modifier=""):
    """Method that corresponds to a DEL request for deleting objects"""
    url = endpoint_url(self.folder, self.endpoint, id=id, modifier=f"{modifier}", conn=self.conn)
    return request_json(url, options={}, payload="", conn=self.conn, request_type="DEL")

  def delete_subitem(self, id, modifier_id, modifier):
    """Method that corresponds to a DEL request for deleting objects with subitems"""
    url = endpoint_url(self.folder, self.endpoint, id=id, modifier=f"{modifier}/{modifier_id}", conn=self.conn)
    return request_json(url, options={}, payload="", conn=self.conn, request_type="DEL")

  def export_one(self, export_endpoint, export_from=""):
    """Export Doc String"""
    url = endpoint_url(self.folder, "export", id="", modifier=f"{export_endpoint}", conn=self.conn)
    return request_json(url, options={"from": export_from}, payload="", conn=self.conn, request_type="GET")

  def export_all(self, export_endpoint, export_from=""):
    """Export All Doc String"""
    counter = 1
    print(f"{export_endpoint} {counter}: {export_from}")
    response = self.export_one(export_endpoint, export_from)
    data = response["data"]
    if data == []: return []
    has_more = response["hasMore"]
    while has_more:
      counter += 1
      export_from = response["continueFrom"]
      print(f"{export_endpoint} {counter}: {export_from}")
      response = self.export_one(export_endpoint, export_from)
      data.extend(response["data"])
      has_more = response["hasMore"]
    print(f"Export Data Complete. {len(data)} rows exported.")
    return data


"""A high-level package for ineracting with the ServiceTitan API v2.

ServicePytan provides an API for easily constructing requests to the ServiceTitan API v2. 
The `Endpoint` class is the primary means to constructing requests based on the organization detailed 
on developer site.

  Resources:
    - https://developer.servicetitan.io

  Examples:
    >>> import servicepytan
    >>> servicetitan_endpoint = servicepytan.Endpoint("jpm","jobs")
    >>> data = servicetitan_endpoint.get_one(id=12345678)
"""

__author__ = """Elliot Palmer"""
__email__ = 'elliot@ecoplumbers.com'
__version__ = '0.3.0'

# MODULES
import requests
import json

# GLOBALS
AUTH_ROOT = "https://auth.servicetitan.io"
URL_ROOT = "https://api.servicetitan.io"

from servicepytan.requests import Endpoint
from servicepytan.reports import Report
from servicepytan.data import DataService
from servicepytan._dates import _convert_date_to_api_format



"""Provides methods for simplified and opininated ways of retrieving data from the ServiceTitan API"""
from servicepytan.requests import Endpoint
from servicepytan._dates import _convert_date_to_api_format
from servicepytan.utils import get_timezone_by_file
class DataService:
  """Primary class for executing data methods.

  The DataService class retrieves the configuration file and authentication settings.
  The methods included are a selection of common data pulls with a simplified API generally
  based on retrieving data between a date range.

  Attributes:
      conn: a dictionary containing the credential config.
  """
  def __init__(self, conn=None):
    """Inits DataService with configuration file and authentication settings."""
    self.conn = conn
    self.timezone = get_timezone_by_file(conn)

  def get_jobs_completed_between(self, start_date, end_date, job_status=["Completed","Scheduled","InProgress","Dispatched"]):
    """Retrieve all jobs completed between the start and end date"""
    data = []
    for status in job_status:
      options = {
        "jobStatus": status,
        "completedOnOrAfter": _convert_date_to_api_format(start_date, self.timezone),
        "completedBefore": _convert_date_to_api_format(end_date, self.timezone)
      }
      data.extend(Endpoint("jpm", "jobs").get_all(options))
    
    return data

  def get_jobs_created_between(self, start_date, end_date):
    """Retrieve all jobs created between the start and end date"""
    options = {
      "createdOnOrAfter": _convert_date_to_api_format(start_date, self.timezone),
      "createdBefore": _convert_date_to_api_format(end_date, self.timezone)
    }
    return Endpoint("jpm", "jobs").get_all(options)

  def get_appointments_between(self, start_date, end_date, appointment_status=["Scheduled", "Dispatched", "Working","Done"]):
    """Retrieve all appointments that start between the start and end date"""
    data = []
    for status in appointment_status:
      options = {
        "status": status,
        "startsOnOrAfter":_convert_date_to_api_format(start_date, self.timezone),
        "startsBefore":_convert_date_to_api_format(end_date, self.timezone)
      }
      data.extend(Endpoint("jpm", "appointments").get_all(options))
    
    return data

  def get_sold_estimates_between(self, start_date, end_date):
    """Retrieve all sold estimates that sold between the start and end date"""
    options = {
        "active": "True",
        "soldAfter":_convert_date_to_api_format(start_date, self.timezone),
        "soldBefore":_convert_date_to_api_format(end_date, self.timezone)
      }
    return Endpoint("sales", "estimates").get_all(options)

  def get_total_sales_between(self, start_date, end_date):
    """Retrieves total sales $ between start and end date"""
    data = self.get_sold_estimates_between(start_date, end_date)
    sales = 0
    for row in data:
      for sku in row["items"]:
        sales += sku['total']
    
    return sales

  def get_purchase_orders_created_between(self, start_date, end_date):
    """Retrieve all purchase orders between the start and end date"""
    options = {
        "createdOnOrAfter":_convert_date_to_api_format(start_date, self.timezone),
        "createdBefore":_convert_date_to_api_format(end_date, self.timezone)
      }
    return Endpoint("inventory", "purchase-orders").get_all(options)

  def get_jobs_modified_between(self, start_date, end_date):
    """Retrieve all jobs modified between the start and end date"""
    options = {
      "modifiedOnOrAfter":_convert_date_to_api_format(start_date, self.timezone),
      "modifiedBefore":_convert_date_to_api_format(end_date, self.timezone)
    }
    data = Endpoint("jpm", "jobs").get_all(options)
    
    return data

  def get_employees(self, active="True"):
    """Retrieve active employee list"""
    options = {
        "active": active
      }
    return Endpoint("settings", "employees").get_all(options)

  def get_technicians(self, active="True"):
    """Retrieve active technician list"""
    options = {
        "active": active
      }
    return Endpoint("settings", "technicians").get_all(options)

  def get_tag_types(self, active="True"):
    """Retrieve active tag-types list"""
    options = {
        "active": active
      }
    return Endpoint("settings", "tag-types").get_all(options)

  def get_business_units(self, active="True"):
    """Retrieve active business units list"""
    options = {
        "active": active
      }
    return Endpoint("settings", "business-units").get_all(options)